MISSING_SETUP = 'Missing implementation of "setup" in PactHelper'
MISSING_TEAR_DOWN = 'Missing implementation of "tear_down" in PactHelper'


class PactHelper(object):
    test_url = 'localhost'
    test_port = 9999



from pact_test.either import *
from pact_test.utils.logger import *
from pact_test.matchers.matcher import *
try:
    import urllib.parse as parse
except ImportError:
    import urllib as parse


def match(actual, expected):
    return _match_path(actual, expected)\
        .concat(_match_query, expected)\
        .concat(_match_method, expected)\
        .concat(_match_headers, expected)\
        .concat(_match_body, expected)


def _match_body(actual, expected):
    expected_body = expected.body
    actual_body = actual.body

    if expected_body is None and actual_body is None:
        return Right(actual)

    if is_string(expected_body) and is_string(actual_body) and expected_body == actual_body:
        return Right(actual)

    if type(expected_body) is dict \
            and type(actual_body) is dict \
            and _match_dicts_all_keys_and_values(expected_body.copy(), actual_body.copy()):
            return Right(actual)

    return Left(build_error_message('body', expected_body, actual_body))


def _match_dicts_all_keys_and_values(d1, d2):
    d1_keys = d1.keys()
    d2_keys = d2.keys()

    all_keys = set(d2_keys).issubset(set(d1_keys))
    all_values = match_dicts_all_values(d1, d2)

    return all_keys and all_values


def _match_headers(actual, expected):
    actual_dict = dict(pair for d in actual.headers for pair in sorted(d.items()))
    expected_dict = dict(pair for d in expected.headers for pair in sorted(d.items()))

    insensitive_actual = {k.upper(): v for (k, v) in sorted(actual_dict.items())}
    insensitive_expected = {k.upper(): v for (k, v) in sorted(expected_dict.items())}

    if is_subset(insensitive_expected, insensitive_actual):
        return Right(actual)
    return Left(build_error_message('headers', expected.headers, actual.headers))


def _match_path(actual, expected):
    if actual.path == expected.path:
        return Right(actual)
    return Left(build_error_message('path', expected.path, actual.path))


def _match_query(actual, expected):
    if _encode(actual.query) == _encode(expected.query):
        return Right(actual)
    return Left(build_error_message('query', expected.query, actual.query))


def _encode(str):
    return parse.quote(str, safe="%3D")


def _match_method(actual, expected):
    if actual.method.upper() == expected.method.upper():
        return Right(actual)
    return Left(build_error_message('method', expected.method.upper(), actual.method.upper()))



import os
import json


def load_acceptance_test(path_to_file):
    acceptance_test_filename = os.path.basename(path_to_file).split('.py')[0]
    acceptance_test_filename += '.json'
    dir_name = os.path.dirname(path_to_file)

    path = os.path.join(dir_name, acceptance_test_filename)
    with open(path) as f:
        data = json.load(f)

    return data



import os
import imp
import inspect
from pact_test.either import *
from pact_test import PactHelper
from pact_test.constants import MISSING_SETUP
from pact_test.constants import MISSING_TEAR_DOWN
from pact_test.constants import EXTEND_PACT_HELPER
from pact_test.constants import MISSING_PACT_HELPER


def load_pact_helper(consumer_tests_path):
    return _path_to_pact_helper(consumer_tests_path)\
               .concat(_load_module, 'pact_helper') \
               >> _load_user_class


def _load_user_class(user_module):
    user_class = None

    for name, obj in inspect.getmembers(user_module):
        if inspect.isclass(obj) and len(inspect.getmro(obj)) > 2 and issubclass(obj, PactHelper):
            user_class = obj()

    if user_class is None:
        return Left(EXTEND_PACT_HELPER)
    if hasattr(user_class, 'setup') is False:
        return Left(MISSING_SETUP)
    if hasattr(user_class, 'tear_down') is False:
        return Left(MISSING_TEAR_DOWN)

    return Right(user_class)


def _load_module(path, module_name):
    try:
        return Right(imp.load_source(module_name, path))
    except Exception:
        return Left(MISSING_PACT_HELPER + path + '".')


def _path_to_pact_helper(consumer_tests_path):
    path = os.path.join(consumer_tests_path, 'pact_helper.py')
    if os.path.isfile(path) is False:
        msg = MISSING_PACT_HELPER + consumer_tests_path + '".'
        return Left(msg)
    return Right(path)


def format_headers(pact):
    for interaction in pact.get('interactions', []):
        req_headers = interaction.get('request').get('headers')
        fixed_req_headers = {}
        for h in req_headers:
            fixed_req_headers.update(h)
        interaction['request']['headers'] = fixed_req_headers

        res_headers = interaction.get('response').get('headers')
        fixes_req_headers = {}
        for h in res_headers:
            fixes_req_headers.update(h)
        interaction['response']['headers'] = fixes_req_headers
    return pact



import os
import json
import copy
from pact_test.either import *
from pact_test.utils.logger import *
from pact_test.config.config_builder import Config
from pact_test.repositories.pact_broker import upload_pact
from pact_test.utils.pact_helper_utils import format_headers
from pact_test.utils.logger import log_consumers_test_results
from pact_test.utils.logger import log_providers_test_results
from pact_test.runners.service_consumers.test_suite import ServiceConsumerTestSuiteRunner
from pact_test.runners.service_providers.test_suite import ServiceProviderTestSuiteRunner


def verify(verify_consumers=False, verify_providers=False):
    config = Config()

    if verify_consumers:
        run_consumer_tests(config)
    if verify_providers:
        run_provider_tests(config)


def run_consumer_tests(config):
    test_results = ServiceConsumerTestSuiteRunner(config).verify()
    log_consumers_test_results(test_results)


def run_provider_tests(config):
    test_results = ServiceProviderTestSuiteRunner(config).verify()
    log_providers_test_results(test_results)
    if type(test_results) is Right:
        write_pact_files(config, copy.deepcopy(test_results.value))
        upload_pacts(config, copy.deepcopy(test_results.value))


def upload_pacts(config, pacts):
    for pact in pacts:
        provider = pact['provider']['name']
        consumer = pact['consumer']['name']
        is_uploadable = all(interaction['status'] == 'PASSED' for interaction in pact['interactions'])
        if is_uploadable:
            ack = upload_pact(provider, consumer, pact, base_url=config.pact_broker_uri)
            msg = 'Pact between ' + pact['consumer']['name'] + ' and ' + pact['provider']['name']
            error(ack.value) if type(ack) is Left else info(msg + ' successfully uploaded')


def write_pact_files(config, pacts):
    pacts_directory = os.path.join(os.getcwd(), config.pacts_path)
    if not os.path.exists(pacts_directory):
        info('Creating Pacts directory at: ' + str(pacts_directory))
        os.makedirs(pacts_directory)

    for pact in pacts:
        pact_payload = format_headers(copy.deepcopy(pact))
        filename = pact_payload['consumer']['name'] + '_' + pact_payload['provider']['name'] + '.json'
        filename = filename.replace(' ', '_').lower()
        info('Writing pact to: ' + filename)
        with open(os.path.join(pacts_directory, filename), 'w+') as file:
            file.write(json.dumps(pact_payload, indent=2))



class PactRequest(object):
    path = ''
    query = ''
    body = None
    headers = []
    method = 'GET'

    def __init__(self, method=None, body=None, headers=None, path=None, query=None):
        self.body = body or self.body
        self.path = path or self.path
        self.query = query or self.query
        self.method = method or self.method
        self.headers = headers or self.headers



from pact_test.models.pact_helper import PactHelper
from pact_test.models.service_consumer_test import state
from pact_test.models.service_consumer_test import pact_uri
from pact_test.models.service_provider_test import given
from pact_test.models.service_provider_test import with_request
from pact_test.models.service_provider_test import has_pact_with
from pact_test.models.service_provider_test import upon_receiving
from pact_test.models.service_provider_test import service_consumer
from pact_test.models.service_consumer_test import honours_pact_with
from pact_test.models.service_provider_test import will_respond_with
from pact_test.models.service_consumer_test import ServiceConsumerTest
from pact_test.models.service_provider_test import ServiceProviderTest


state = state
given = given
pact_uri = pact_uri
PactHelper = PactHelper
with_request = with_request
has_pact_with = has_pact_with
upon_receiving = upon_receiving
service_consumer = service_consumer
will_respond_with = will_respond_with
honours_pact_with = honours_pact_with
ServiceConsumerTest = ServiceConsumerTest
ServiceProviderTest = ServiceProviderTest
